#include "CBirds.h"
#include "RwV3D.h"
#include "RwRGBA.h"
#include "RenderWare.h"
#include "CColLine.h"
#include "ColSphere.h"
#include "Addresses.h"
#include <cmath>

#define BIRD_VERTEX_COUNT	8

#pragma warning(disable : 4482)

CBird* CBirds::BirdPool = nullptr;
CBird* CBirds::PoolEnd = nullptr;
RwV3D* CBirds::LastKilledBirdPos = nullptr;
bool* CBirds::HasKilledAnyBird = nullptr;
int* CBirds::nBirdsCreated = nullptr;
float* CBirds::XOffset;
float* CBirds::YOffset;
float* CBirds::ZOffset;

static int BirdTXD = 0;
RwTexture* HDBirdTex = nullptr;
static const char* HDBirdTexName = "hdbird";

/*#########################################################################################*/

static void LoadBirdTexture()
{
	chdirProgramDir("");
	if(!BirdTXD) BirdTXD = txdAdd("birds");
	txdLoad(BirdTXD, "MODELS\\BIRDS.TXD");
	txdAddRef(BirdTXD);
	txdSetCurrent(BirdTXD);
	HDBirdTex = RwReadTexture(HDBirdTexName, nullptr);
}

static void OnInitBirds()
{
	// Patch indices!
	unsigned int* indices = (unsigned int*)0x8D52F8;
	indices[27] = 8;
	indices[24] = 8;

	LoadBirdTexture();
	WriteMemory((void*)(0x712DE7+2), (DWORD)&HDBirdTex, true);
}

static void OnShutdownBirds()
{
	RwTextureDestroy(HDBirdTex);
	texDictionaryRemove(BirdTXD);
}
/*##########################################################################################*/



void __cdecl CBirds::Init()
{
	for(CBird* bird = BirdPool; bird != PoolEnd; ++bird)
		bird->HasBirdHere = false;
	*nBirdsCreated = 0;
	*HasKilledAnyBird = false;

	//--
	OnInitBirds();
}

void __cdecl CBirds::Shutdown()
{
	for(CBird* bird = BirdPool; bird != PoolEnd; ++bird)
		bird->HasBirdHere = false;
	*nBirdsCreated = 0;

	//--
	OnShutdownBirds();
}


void __cdecl CBirds::CheckIfHitByFire(RwV3D* p1, RwV3D* p2)
{
	ColSphere BirdSphere;	// Bird Bounding Sphere
	CColLine Fire;			// Weapon Fire line

	for(CBird* bird = BirdPool; bird != PoolEnd; ++bird)
	{
		if(bird->HasBirdHere)
		{
			CColLine__Constructor(&Fire, 0, p1, p2);
			BirdSphere.sphere.center = bird->pos;
			BirdSphere.sphere.radius = 0.5;
			if(CColLine__HasCollisionWithSphere(&Fire, &BirdSphere))
			{
				*LastKilledBirdPos = bird->pos;
				*nBirdsCreated -= 1;
				bird->HasBirdHere = false;
			}
		}
	}
}

void __cdecl CBirds::Process()
{
	RwV3D CamPos;

	// Get camera position
	if(camera->__parent.m_pCoords)
		CamPos = (reinterpret_cast<Placement*>(((char*)camera->__parent.m_pCoords + 0x30))->pos);
	else
		CamPos = camera->__parent.placement.pos;

	// Only create the bird if active interior is 0, has free slots in bird pool, hours is between 5~22,
	// and a bitwise operation with playingTime
	if(*activeInterior == 0 && *nBirdsCreated < 6
		&& *currentTimeHours < 22 && *currentTimeHours > 5 && (*playingTime & 0x1FF) == 6)
	{
		int birdType;
		int birdCount;
		float randHeight;
		float distance;
		long double angle;
		float waterLevel;
		CVector v;
		RwV3D BirdPos, BirdDest;

		// Get number of birds in group
		birdCount = random(1, 7 - *nBirdsCreated);

		// Get bird type...
		//	If camera is near water, bird type is 0;
		//	If camera is not near water and in desert, bird type is 1
		//	If camera is not near water and not in desert, bird type is 2
		if(camera->DistanceToWater >= 30.0f)
		{
			if(*wTownID == TownID::Desert)
			{
				birdType = 1;
				if(birdCount >= 3 - *nBirdsCreated)
					birdCount = 3 - *nBirdsCreated;
			}
			else birdType = 2;
		}
		else birdType = 0;

		if(birdCount > 0)
		{
			rand(); rand();
			
			// Get bird height and distance from camera
			switch(birdType)
			{
				case 0:
					randHeight = randomf(4.0, 13.0);
					distance = 45;//4.5E1;
					break;

				case 1:
					randHeight = randomf(15.0, 25.0);
					distance = 80;//8.0E1;
					break;

				case 2:
					randHeight = randomf(2.0, 10.0);
					distance = 40;//4.0E1;
					break;
			}

			// Get bird height, if too slow don't continue
			randHeight = randHeight + CamPos.z;
			if(randHeight > 5.0f)
			{
				// Get bird "angle"
				if(rand() & 1)
				{
					angle = (long double)(rand() & 0xFF) * 0.02453125081956387;
				}
				else
				{
					v.x = camera->field_974.matrix.top.x;
					v.y = camera->field_974.matrix.top.y;
					v.z = 0;
					if(camera->field_974.matrix.top.x == 0.0)
						v.x = (float)(9.99999977648258209228515625e-3);
					CVector__normalize(&v);
					angle = (long double)((rand() & 0xFF) - 128) * 0.00390625 + atan2(v.y, v.x);
				}

				// Now we have bird position
				BirdPos.x = float( sin(angle) * distance + CamPos.x );
				BirdPos.y = float( cos(angle) * distance + CamPos.y );
				BirdPos.z = float( randHeight );

				// Check if the position is not in the water
				if(!CWaterLevel__GetWaterLevelNoWaves(BirdPos.x, BirdPos.y, BirdPos.z, &waterLevel, nullptr, nullptr)
				|| waterLevel + 4.0 < BirdPos.z)
				{
					// Get bird destination

					v.x = camera->field_974.matrix.top.x;
					v.y = camera->field_974.matrix.top.y;
					v.z = 0;
					CVector__normalize(&v);

					BirdDest.x = (float)(CamPos.x + (v.x * 8.0));
					BirdDest.y = (float)(CamPos.y + (v.y * 8.0));
					BirdDest.z = BirdPos.z;

					// Add birds to pool!
					CBirds::Add(BirdPos, BirdDest, birdCount, birdType, true);
				}
			}

		}
	}

	// Remove birds out of vision
	{
		CBird* bird = &CBirds::BirdPool[ (*playingTime % CBirds::PoolSize)  ];
		if(bird->HasBirdHere)
		{
			float vx = bird->pos.x - CamPos.x;
			float vy = bird->pos.y - CamPos.y;

			if(  (bird->MaxBirdDistance * bird->MaxBirdDistance) < ((vy * vy) + (vx * vx)) )
			{
				*nBirdsCreated -= 1;
				bird->HasBirdHere = false;
			}

		}
	}

	// Update birds position
	{
		for(CBird* bird = CBirds::BirdPool; bird != CBirds::PoolEnd; ++bird)
		{
			if(bird->HasBirdHere && bird->field_38 && *currentTime > (DWORD)bird->field_28)
			{
				if(bird->MustDoCurves)	//
				{
					long double v19 = *TimeStep * 0.002000000094994903;
					long double v20 = bird->Velocity.x;
					long double v21 = bird->Velocity.y;
					long double v22 = sin(v19);
					long double v23 = cos(v19);

					bird->Velocity.x = (float)(v23 * v20 + v22 * v21);
					bird->Velocity.y = (float)(v23 * v21 - v22 * v20);
				}

				long double v25 = *TimeStep * 0.01999999955296516;

				long double v24 = 0.5 * v25;
				if(v24 > 1.0f) v24 = 1.0f;

				bird->field_C.x = (float)( (1.0 - v24) * bird->field_C.x + bird->Velocity.x * v24 );
				bird->field_C.y = (float)( (1.0 - v24) * bird->field_C.y + bird->Velocity.y * v24 );
				bird->field_C.z = (float)( (1.0 - v24) * bird->field_C.z + bird->Velocity.z * v24 );

				bird->pos.x += (float)(v25 * bird->field_C.x);
				bird->pos.y += (float)(v25 * bird->field_C.y);
				bird->pos.z += (float)(v25 * bird->field_C.z);

				bird->angle = atan2(bird->Velocity.x, bird->Velocity.y);
														// WHAT?,
														// Don't know why x is in the first param ans y in the second param!!
														// but if I do this in the usual way the bird will have a wrong angle

			}

		}
	}


}


int __cdecl CBirds::Add(RwV3D pos, RwV3D dest, int numBirds, int birdType, bool checkObstacles)
{
	if(!numBirds) return 0;

	float MaxBirdDistance;
	int i = -1;

	switch(birdType)
	{
		case 0:
			MaxBirdDistance = 45;
			break;

		case 1:
			MaxBirdDistance = 80;
			break;

		case 2:
			MaxBirdDistance = 40;
			break;

		default:
			birdType = 0;
			MaxBirdDistance = 45;
			break;
	}

	for(int count = 0; count < numBirds; ++count)
	{
		// Optimized: No need to reloop in the pool, continue from where I've stopped
		++i;
		while(&BirdPool[i] < PoolEnd)
		{
			if(!BirdPool[i].HasBirdHere) break;
			++i;
		}

		if(&BirdPool[i] >= PoolEnd) return count;	// No free slot on the pool
		
		CBird* bird = &BirdPool[i];
		CVector v;

		bird->pos = pos;
		
		v.x = dest.x - bird->pos.x;
		v.y = dest.y - bird->pos.y;
		v.z = dest.z - bird->pos.z;
		CVector__normalize(&v);

		dest.x = (float)(v.x * MaxBirdDistance * 2.400000095367432 + bird->pos.x);
		dest.y = (float)(v.y * MaxBirdDistance * 2.400000095367432 + bird->pos.y);
		dest.z = (float)(v.z * MaxBirdDistance * 2.400000095367432 + bird->pos.z);

		if(checkObstacles
		&& !IsLineOfSightClear(&bird->pos, &dest, true, false, false, false, false, false, false))
			continue;

		int randSpeed = rand() % 31;

		bird->HasBirdHere = true;

		bird->angle = atan2(v.y, v.x);
		bird->MaxBirdDistance = MaxBirdDistance;

		bird->field_38 = true;
		bird->field_28 = 0;
		bird->MustDoCurves = false;

		bird->pos.x += XOffset[count];
		bird->pos.y += YOffset[count];
		bird->pos.z += ZOffset[count];

		//bird->WingSpeed = 1000 - 12 * randSpeed;	-- USELESS, SINCE THIS WILL BE DEFINED IN THE SWITCH BELOW

		switch(birdType)
		{
			case 0:
			{
				long double v36 = (long double)randSpeed * 0.01999999955296516 + 4.0;

				bird->Velocity.x = (float)(v.x * v36);
				bird->Velocity.y = (float)(v.y * v36);
				bird->Velocity.z = (float)(v.z * v36);

				bird->field_C = bird->Velocity;

				// ? Related to wings
				bird->WingSpeed = 1000 - 12 * randSpeed;

				// Bird size
				// randomf(0.3999999761581421, 0.800000011920929)
				bird->Size = (float)( (long double)rand() * 0.00003051850944757462 * 0.4000000357627869 + 0.800000011920929 );

				// Polygons color
				bird->PolyColor1 = bird->PolyColor2 = bird->PolyColor3 = (rand() & 0x3F) + 80;
				bird->PolyColor6 = bird->PolyColor5 = bird->PolyColor4 = (rand() & 0x3F) - 76;
				break;
			}
				

			case 1:
			{
				long double v28 = (long double)randSpeed * 0.01999999955296516 + 3.0;

				bird->Velocity.x = (float)(v.x * v28);
				bird->Velocity.y = (float)(v.y * v28);
				bird->Velocity.z = (float)(v.z * v28);

				bird->field_C = bird->Velocity;

				// bird->Size = randomf(1.5, 2.0) - bird size
				bird->Size = float((long double)rand() * 0.00003051850944757462 * 0.5 + 2.0);
				
				// ? Related to wing
				if(rand() & 4)
					bird->WingSpeed = 1000000;
				else
					bird->WingSpeed = 12 * (125 - randSpeed);

				// Must do curves?
				if(rand() & 8)
					bird->MustDoCurves = true;

				// Polygons color
				bird->PolyColor1 = 30;
				bird->PolyColor2 = 15;
				bird->PolyColor3 = 10;
				bird->PolyColor4 = 80;
				bird->PolyColor5 = 15;
				bird->PolyColor6 = 10;

				break;
			}

			case 2:
			{
				long double v19 = (long double)randSpeed * 0.01999999955296516 + 5.0;
				
				bird->Velocity.x = (float)(v.x * v19);
				bird->Velocity.y = (float)(v.y * v19);
				bird->Velocity.z = (float)(v.z * v19);

				bird->field_C = bird->Velocity;

				// ? wing
				bird->WingSpeed = 500 - 6 * randSpeed;

				// bird size
				// randomf(0.3999999761581421, 0.5)
				bird->Size = (float)((long double)rand() * 0.00003051850944757462 * 0.1000000238418579 + 0.5);

				bird->PolyColor3 = bird->PolyColor2 = bird->PolyColor1 = (rand() & 0x7F) + -128;
				bird->PolyColor6 = bird->PolyColor5 =  bird->PolyColor4 = (rand() & 0x7F) + -128;

				break;
			}

		}

		*CBirds::nBirdsCreated += 1;
	}


	return numBirds;
}


void __cdecl CBirds::Render()
{
	if(!*CBirds::nBirdsCreated) return;

	CMatrix m;
	int v4, v58, v56;
	float v55, v52, v59, v60;
	float a3, a4;
	RwV3D outPoint;
	RwV3D point;
	RwRGBA polyColor;
	unsigned char birdAlphaLevel;
	RxObjSpace3dVertex* pVerts; pVerts = (RxObjSpace3dVertex*)(0xC4D958);

	*numIndices = 0;
	*numVerts = 0;

	v4 = 0, v58 = 0, v56 = 0;

	for(CBird* bird = CBirds::BirdPool; bird != CBirds::PoolEnd; ++bird)
	{
		if(bird->HasBirdHere)
		{
			m.matrixPtr = 0;
			m.haveRwMatrix = 0;

			CMatrix__initIdentityMatrixAtPoint(&m, 0, bird->pos);

			if(bird->field_38 == 1 || bird->field_38 == 2)	// Always true???
			{
				m.matrix.pos.z += (float)
					sin((long double)((unsigned int)(v4 + *currentTime) % bird->WingSpeed) *
					6.28000020980835 / (long double)bird->WingSpeed) * 0.1000000014901161;
			}

			if(ConvertGlobalXYToScreenXY(&m.matrix.pos, &outPoint, &a3, &a4, 0, 1))
			{
				float angle = bird->angle;

				m.matrix.at.z = bird->Size;
				m.matrix.at.x = m.matrix.at.y = 0;
				m.matrix.right.z = m.matrix.top.z = 0;

				m.matrix.right.x = m.matrix.top.y = cos(angle) * bird->Size;
				m.matrix.right.y = m.matrix.top.x = sin(angle) * bird->Size;
				m.matrix.right.y = -m.matrix.right.y;

				if(bird->field_38 == 1 || bird->field_38 == 2)	// Always true???
				{

					// Setup indices
					{
						int* birdIndices = (int*)0x8D52F8;	// Bird indices
						short* rIndices = (short*)((char*)0xC4B958  + (2 * *numIndices));

						for(int i = 0; i < 30; ++i)
						{
							rIndices[i + 0] = *numVerts + birdIndices[i];
							rIndices[i + 30] = *numVerts + birdIndices[i] + BIRD_VERTEX_COUNT;
						}
					}

					// Setup wings
					{
						long double v17 = 6.28000020980835 / (long double)bird->WingSpeed;

						long double v51 = (-sin(((*currentTime + v4 + bird->WingSpeed / 6) % bird->WingSpeed) * v17))
							/ (bird->Size >= 1.0? bird->Size : 1.0);

						long double v20 = (-sin(v17 * (long double)((v4 + *currentTime) % bird->WingSpeed)))
							/ (bird->Size >= 1.0? bird->Size : 1.0);

						v55 = (float)(cos(v51) * 0.5);
						v52 = (float)(sin(v51) * 0.5);
						v59 = (float)(cos(v20) * 0.5);
						v60 = (float)(sin(v20) * 0.5);
					}

					// Get bird alpha level!
					{
						RwV3D* CamPos;
						if(camera->__parent.m_pCoords)
							CamPos = &(reinterpret_cast<Placement*>(((char*)camera->__parent.m_pCoords + 0x30))->pos);
						else
							CamPos = &camera->__parent.placement.pos;

						auto v25 = bird->pos.z - CamPos->z;
						auto v26 = bird->pos.y - CamPos->y;
						auto v27 = bird->pos.x - CamPos->x;

						auto v24 = sqrt(v27 * v27 + v26 * v26 + v25 * v25);
						auto v23 = bird->MaxBirdDistance * 0.699999988079071;

						if (v24 > v23)
						{
							auto v28 = 1.0 - (v24 - v23) / (bird->MaxBirdDistance * 0.300000011920929);
							if(v28 < 0.0)
								v28 = 0.0;
							birdAlphaLevel = (unsigned char)(v28 * 255.0);
						}
						else birdAlphaLevel = 255;
					}

					polyColor.alpha = birdAlphaLevel;

					for(int i = 0; i < BIRD_VERTEX_COUNT; ++i)
					{
						switch(i)
						{
									//  .
							case 0:	// /|  -- replaced with vertex 8!! Now it's just used for the wing
							case 1:	//     -- bottom wing vertex in the middle of the bird body
							case 7: // /|. -- right bottom vertex of the bird body
								point.x = 0;
								point.z = 0;
								polyColor.red = bird->PolyColor1;
								polyColor.green = bird->PolyColor2;
								polyColor.blue = bird->PolyColor3;
								break;

							case 6:	// ./| -- left bottom of the bird body
								point.x = (float)1.00000001490116119384765625E-1;
								point.z = 0;
								polyColor.red = bird->PolyColor1;
								polyColor.green = bird->PolyColor2;
								polyColor.blue = bird->PolyColor3;
								break;

						// Wing vertexes are below

							case 2:
							case 3:
								point.x = v55;
								point.z = v52;
								polyColor.red = bird->PolyColor1;
								polyColor.green = bird->PolyColor2;
								polyColor.blue = bird->PolyColor3;
								break;

							case 4:
							case 5:
								point.x = v59 + v55;
								point.z = v60 + v52;
								polyColor.red = bird->PolyColor4;
								polyColor.green = bird->PolyColor5;
								polyColor.blue = bird->PolyColor6;
								break;
						}

						if(i != 8) point.y = *((float*)0x8D52D8 + i);
						else point.y = *((float*)0x8D52D8 + 0);

						int vidx = (i + *numVerts);
						RxObjSpace3dVertex* LeftVert = &pVerts[0 + vidx];
						RxObjSpace3dVertex* RightVert = &pVerts[BIRD_VERTEX_COUNT + vidx];

						// Setup vertex position
						transformPoint(&outPoint, &m, &point);
						RwIm3DVertexSetPos(LeftVert, outPoint.x, outPoint.y, outPoint.z);
						point.x = -point.x;	// The other side
						transformPoint(&outPoint, &m, &point);
						RwIm3DVertexSetPos(RightVert, outPoint.x, outPoint.y, outPoint.z);

						// Setup vertex RGBA
						RwIm3DVertexSetRGBA(LeftVert, polyColor.red, polyColor.green,
							polyColor.blue, polyColor.alpha);
						RwIm3DVertexSetRGBA(RightVert, BYTE(polyColor.red * 0.8), BYTE(polyColor.green * 0.8),
							BYTE(polyColor.blue * 0.8), polyColor.alpha );
						
						// Setup texture coordinates (UV)
						RwIm3DVertexSetU(LeftVert, *((float*)0x8D52B8 + i) );
						RwIm3DVertexSetV(LeftVert, *((float*)0x8D5298 + i) );
						RwIm3DVertexSetU(RightVert, *((float*)0x8D52B8 + i) );
						RwIm3DVertexSetV(RightVert, *((float*)0x8D5298 + i) );

					}

					v4 = v56;
					*numIndices += 60;
					*numVerts += BIRD_VERTEX_COUNT*2;
				}


			}
			CMatrix__destructor(&m);
		}

		v4 += 100;
		v56 = v4;
	}

	if(*numIndices != 0)
	{
		void (*sub_722150)() = (decltype(sub_722150))( 0x722150 );
		
		RwRenderStateSet(rwRENDERSTATESRCBLEND, (void*) 5);
		RwRenderStateSet(rwRENDERSTATEDESTBLEND, (void*) 6);
		RwRenderStateSet(rwRENDERSTATETEXTUREADDRESS, (void*) 3);
		RwRenderStateSet(rwRENDERSTATEFOGENABLE, (void*) 0);
		RwRenderStateSet(rwRENDERSTATEVERTEXALPHAENABLE, (void*) 1);
		RwRenderStateSet(rwRENDERSTATETEXTURERASTER, (void*) HDBirdTex->raster);

		sub_722150();
	}
}
